from abc import ABC, abstractmethod
from functools import cached_property, wraps
from typing import Any, Callable

import numpy as np
from scipy import special


def optionally_cached_property(func: Callable) -> property:
    cached = cached_property(func)

    @wraps(func)
    def wrapper(cls: Any) -> Any:
        if getattr(cls, "_cache_properties", False):
            if not hasattr(cached, "__wrapped__"):
                cached.__set_name__(cls, func.__name__)
            return cached.__get__(cls)
        return func(cls)

    return property(wrapper)


class StatsBase(ABC):
    @optionally_cached_property
    @abstractmethod
    def residuals(self) -> np.ndarray: ...

    @optionally_cached_property
    @abstractmethod
    def fitted_values(self) -> np.ndarray:
        """Estimated y-hat values of training data, after applying inverse link func."""
        ...

    @optionally_cached_property
    @abstractmethod
    def standard_residuals(self) -> np.typing.ArrayLike: ...

    @optionally_cached_property
    @abstractmethod
    def cooks_d(self) -> np.typing.ArrayLike: ...

    @optionally_cached_property
    @abstractmethod
    def leverage(self) -> np.typing.ArrayLike: ...

    @optionally_cached_property
    @abstractmethod
    def parameter_count(self) -> int: ...

    @optionally_cached_property
    def sqrt_abs_residuals(self) -> np.ndarray:
        return np.sqrt(np.abs(self.standard_residuals))

    @optionally_cached_property
    def normalized_quantiles(self) -> np.ndarray:
        val_count = len(self.fitted_values)
        positions = (np.arange(1.0, val_count + 1)) / (val_count + 1.0)
        # ndtri is used under the hood by ppf, but is much faster. It skips validation
        # and processing the loc and scale arguments, which are not needed here.
        return special.ndtri(positions)



import linearmodels
import numpy as np

from lmdiag.statistics.base import StatsBase, optionally_cached_property


class LinearmodelsStats(StatsBase):
    def __init__(
        self,
        lm: linearmodels.iv.results.OLSResults,
        cache: bool = True,
    ) -> None:
        super().__init__()
        self._lm = lm
        self._cache_properties = cache

    @optionally_cached_property
    def residuals(self) -> np.ndarray:
        return self._lm.resids.values

    @optionally_cached_property
    def fitted_values(self) -> np.ndarray:
        return self._lm.fitted_values.squeeze().values

    @optionally_cached_property
    def standard_residuals(self) -> np.ndarray:
        residuals = self.residuals
        h_ii = self.leverage
        var_e = np.sqrt(self._lm.resid_ss / (self._lm.nobs - self._lm.df_model))
        standard_error = var_e * np.sqrt(1 - h_ii)
        return residuals / standard_error

    @optionally_cached_property
    def cooks_d(self) -> np.ndarray:
        h_ii = self.leverage
        cooks_d2 = self.standard_residuals**2 / self.parameter_count
        cooks_d2 *= h_ii / (1 - h_ii)
        return cooks_d2

    @optionally_cached_property
    def leverage(self) -> np.ndarray:
        X = self._lm.model._x
        XtX_inv = np.linalg.inv(np.dot(X.T, X))
        h_ii = np.einsum("ij,jk,ik->i", X, XtX_inv, X)
        return h_ii

    @optionally_cached_property
    def parameter_count(self) -> int:
        """Number of parameters in the model, including intercept.

        Also known as the degrees of freedom of the model.
        """
        return self._lm.df_model



import warnings
from typing import Any, Optional

import numpy as np
from statsmodels.genmod.generalized_linear_model import GLMResults
from statsmodels.regression.linear_model import RegressionResultsWrapper
from statsmodels.robust.robust_linear_model import RLMResults

from lmdiag.statistics.base import StatsBase

try:
    import sklearn
except ImportError:
    sklearn = None

try:
    import linearmodels
except ImportError:
    linearmodels = None


def _warn_x_y() -> None:
    warnings.warn(
        "`x` and `y` arguments are ignored for this model type. Do not pass them.",
        stacklevel=3,
    )


def _init_linearmodels_stats(lm: Any) -> StatsBase:
    from lmdiag.statistics.linearmodels_stats import LinearmodelsStats

    return LinearmodelsStats(lm)


def _init_sklearn_stats(lm: Any, x: np.ndarray, y: np.ndarray) -> StatsBase:
    from lmdiag.statistics.sklearn_stats import SklearnStats

    return SklearnStats(lm, x=x, y=y)


def _init_statsmodels_stats(lm: Any) -> StatsBase:
    from lmdiag.statistics.statsmodels_stats import StatsmodelsStats

    return StatsmodelsStats(lm)


def get_stats(
    lm: Any, x: Optional[np.ndarray] = None, y: Optional[np.ndarray] = None
) -> StatsBase:
    """Gather statistics depending on linear model type."""
    if isinstance(lm, (RegressionResultsWrapper, GLMResults, RLMResults)):
        if x or y:
            _warn_x_y()
        model_stats = _init_statsmodels_stats(lm)

    elif linearmodels and isinstance(
        lm, (linearmodels.iv.results.OLSResults, linearmodels.iv.results.IVResults)
    ):
        if x or y:
            _warn_x_y()
        model_stats = _init_linearmodels_stats(lm)

    elif sklearn and isinstance(lm, sklearn.linear_model.LinearRegression):
        if x is None or y is None:
            raise ValueError("x and y args must be provided this model type!")
        model_stats = _init_sklearn_stats(lm, x, y)

    else:
        raise TypeError(
            "Model type not (yet) supported. Currently supported are linear "
            "models from `statsmodels`, `linearmodels` and `sklearn` packages."
        )

    return model_stats



from typing import TypedDict, Union


class MplKwargs(TypedDict, total=False):
    """Types for the most common matplotlib arguments to please mypy."""

    alpha: Union[float, None]
    c: Union[str, tuple]
    color: Union[str, tuple]
    edgecolors: Union[str, tuple]
    fillstyle: str
    fontsize: int
    fontweight: str
    linestyle: str
    linewidth: float
    marker: str
    markeredgecolor: Union[str, tuple]
    markerfacecolor: Union[str, tuple]
    markersize: float
    visible: bool
    zorder: float
    pad: float
    h_pad: float
    w_pad: float
    figsize: tuple


scatter: MplKwargs = {"alpha": 0.5, "color": "C0"}
plot: MplKwargs = {"color": "C1"}
plot_contour: MplKwargs = {"color": "dimgrey", "linestyle": "dashed"}
xy_label: MplKwargs = {}
annotate: MplKwargs = {"color": plot["color"]}
title: MplKwargs = {}
tight_layout: MplKwargs = {"h_pad": 2.5, "w_pad": 3}
subplots: MplKwargs = {}


def use(style: str) -> None:
    """Set predefined style for plots.

    Available styles:
    - 'black_and_red' (mimics style of R's lm.diag)

    Args:
        style: Name of the preset style.

    Raises:
        ValueError: If style is unknown.
    """
    if style == "black_and_red":
        scatter.update(
            {"marker": "o", "color": "none", "edgecolors": "black", "linewidth": 1}
        )
        plot.update({"color": "red"})
        plot_contour.update({"color": "dimgrey"})
        annotate.update({"color": plot["color"]})
        subplots.update({"figsize": (10, 7)})
        title.update({"fontsize": 15})
    else:
        raise ValueError(f"Unknown style '{style}'.")



import itertools
from typing import Callable

import numpy as np
import pytest

from lmdiag.statistics.linearmodels_stats import LinearmodelsStats
from lmdiag.statistics.sklearn_stats import SklearnStats
from lmdiag.statistics.statsmodels_stats import StatsmodelsStats


@pytest.mark.parametrize("x_dims", [1, 3, 5])
@pytest.mark.parametrize(
    "attr",
    [
        "residuals",
        "fitted_values",
        "standard_residuals",
        "cooks_d",
        "leverage",
        "parameter_count",
    ],
)
def test_lm_stats_modules(
    attr: str,
    statsmodels_factory: Callable,
    linearmodels_factory: Callable,
    sklearn_factory: Callable,
    x_dims: int,
) -> None:
    model_stats_to_compare = [
        StatsmodelsStats(statsmodels_factory(x_dims=x_dims)),
        LinearmodelsStats(linearmodels_factory(x_dims=x_dims)),
        SklearnStats(*sklearn_factory(x_dims=x_dims)),
    ]

    for stats_a, stats_b in itertools.combinations(model_stats_to_compare, 2):
        vector_a = getattr(stats_a, attr)
        vector_b = getattr(stats_b, attr)

        assert type(vector_a) == type(vector_b), (
            attr,
            stats_a.__class__.__name__,
            stats_b.__class__.__name__,
        )

        assert np.allclose(vector_a, vector_b, atol=1e-10), (
            attr,
            stats_a.__class__.__name__,
            stats_b.__class__.__name__,
        )



import numpy as np
from sklearn.linear_model import LinearRegression

from lmdiag.statistics.base import StatsBase, optionally_cached_property


class SklearnStats(StatsBase):
    def __init__(
        self,
        lm: LinearRegression,
        x: np.ndarray,
        y: np.ndarray,
        cache: bool = True,
    ) -> None:
        super().__init__()
        self._lm = lm
        self._X = x
        self._y = y
        self._cache_properties = cache

    @optionally_cached_property
    def residuals(self) -> np.ndarray:
        return self._y - self.fitted_values

    @optionally_cached_property
    def fitted_values(self) -> np.ndarray:
        return self._lm.predict(self._X)

    @optionally_cached_property
    def standard_residuals(self) -> np.ndarray:
        residuals = self.residuals
        h_ii = self.leverage
        resid_ss: np.ndarray = np.sum(residuals**2)
        nobs = len(self._y)
        df_model = self.parameter_count
        var_e = np.sqrt(resid_ss / (nobs - df_model))
        standard_error = var_e * np.sqrt(1 - h_ii)
        return residuals / standard_error

    @optionally_cached_property
    def cooks_d(self) -> np.ndarray:
        h_ii = self.leverage
        cooks_d2 = self.standard_residuals**2 / self.parameter_count
        cooks_d2 *= h_ii / (1 - h_ii)
        return cooks_d2

    @optionally_cached_property
    def leverage(self) -> np.ndarray:
        X = self._X
        # add constant, like sm.add_constant() does in linearmodels' leverage()
        X = np.column_stack((np.ones(X.shape[0]), X))
        XtX_inv = np.linalg.inv(np.dot(X.T, X))
        h_ii = np.einsum("ij,jk,ik->i", X, XtX_inv, X)
        return h_ii

    @optionally_cached_property
    def parameter_count(self) -> int:
        params_count = len(self._lm.coef_)
        if self._lm.fit_intercept:
            params_count += 1
        return params_count



"""Module for printing descriptions to aid interpretation of the different plots."""

import textwrap
from collections.abc import Iterable
from typing import Literal, Union

plot_descriptions = {
    "resid_fit": {
        "Name": "Residuals vs. Fitted",
        "Method": "lmdiag.resid_fit(lm)",
        "x-Axis": (
            "Fitted Values (Predicted y-hat from training data. If a link function "
            "exists, its inverse is applied)"
        ),
        "y-Axis": (
            "Residuals (The 'error' of the model; "
            "Distances of the y-hat values to the fitted regression line)"
        ),
        "Description": (
            "It's purpose is to identify non-linear patterns in the residuals. "
            "If you see a horizontal red line and the points spread around it without "
            "a recognizable pattern, chances are good, that there is no non-linear "
            "relationship in the data. "
            "If you can see clear pattern or a curve, a linear model might not be the "
            "best choice."
            "The red labels show the indices of three observations with the highest "
            "absolute residuals."
        ),
    },
    "q_q": {
        "Name": "Normal Q-Q",
        "Method": "lmdiag.q_q(lm)",
        "x-Axis": "Theoretical Quantiles (Quantiles from the Normal Distribution)",
        "y-Axis": (
            "Standardized residuals (Quantiles of the values of "
            "the dependent variable in sorted order)"
        ),
        "Description": (
            "It's purpose is to check, if the residuals are following a normal "
            "distribution. "
            "It's good, if the points are aligned on the dashed line. "
            "If only a few points are off, take a look at the other plots. "
            "If lots of points do not follow the line, your distribution might be off "
            "normal, e.g. regarding skew, tails or modality."
        ),
    },
    "scale_loc": {
        "Name": "Scale-Location",
        "Method": "lm.scale_loc(lm)",
        "x-Axis": (
            "Fitted Values (Predicted y-hat from training data. If a link function "
            "exists, its inverse is applied)"
        ),
        "y-Axis": "Square root of the absolute value of the Standardized Residuals.",
        "Description": (
            "It's purpose is to check 'homoscedasticity', the assumption of equal "
            "variance. "
            "The plot shows, if the residuals are spread equally across the range of "
            "predictors (fitted values). "
            "The red line should be horizontal and the scatter points equally spread "
            "in a random matter. "
            "The red labels are the indices of the observations with the highest "
            "absolute residuals."
        ),
    },
    "resid_lev": {
        "Name": "Residuals vs. Leverage",
        "Method": "lmdiag.resid_lev(lm)",
        "x-Axis": (
            "Leverage (The 'influence' of an observation. "
            "A measure of how far away the dependent variable's value of an "
            "observation is from those of other observations.)"
        ),
        "y-Axis": (
            "Residuals (The 'error' of the model; "
            "Distances of the y values to the fitted regression line)"
        ),
        "dashed-Lines": "Cook's Distance, 0.5 (inner) and 1 (outer).",
        "Description": (
            "It's purpose is to identify observations with high influence on "
            "calculating the regression. "
            "Those observation might but not have to be outliers, they are just "
            "extreme cases concerning the regression. "
            "The pattern of the scatter points is not relevant here: interesting are "
            "observations in the top right and bottom right of the plot. "
            "If we have cases outside the Cook's Distance (dashed lines), "
            "removing those would have an high impact on our regression line. "
            "The red labels are the indices of the most influential observations."
        ),
    },
}


# Note on extensive type hints: for some reason, type aliases are not revealed at least
# in VSCode function popovers. Therefore, the duplication seems useful.
# ONHOLD: Verify if this is still true, especially with LiteralString in 3.11.
def _print_desc(
    method: Literal["resid_fit", "q_q", "scale_loc", "resid_lev", "plot"],
) -> None:
    """Print description of plot nicely formatted."""
    for key, val in plot_descriptions[method].items():
        wrapper = textwrap.TextWrapper(
            initial_indent=f"{key:>12}: ", width=79, subsequent_indent=" " * 14
        )
        print(wrapper.fill(val), end="\n\n")  # noqa: T201


def help(  # noqa: A001 # shadowing built-in
    method: Union[
        Iterable[Literal["resid_fit", "q_q", "scale_loc", "resid_lev", "plot"]],
        Literal["resid_fit", "q_q", "scale_loc", "resid_lev", "plot"],
    ] = "plot",
) -> None:
    """Prints description(s) of specified plot method(s) to aid interpretation.

    Args:
        method: The name(s) of the plot method(s) for which to print descriptions.
            Defaults to `"plot"`, which prints descriptions of all four plots.
    """
    if isinstance(method, str):
        method = (
            ["resid_fit", "q_q", "scale_loc", "resid_lev"]
            if method == "plot"
            else [method]
        )

    for m in method:
        if m in plot_descriptions:
            _print_desc(m)
        else:
            print(  # noqa: T201
                f"Unknown plotting method '{method}'."
                "Run lmdiag.help() to print all available method descriptions."
            )



from pathlib import Path
from typing import Callable

import matplotlib.pyplot as plt
import pytest

import lmdiag


@pytest.mark.parametrize(
    "lm_type",
    ["statsmodels", "linearmodels", "sklearn"],
)
def test_plot_generates_expected_image(
    lm_type: str,
    statsmodels_factory: Callable,
    linearmodels_factory: Callable,
    sklearn_factory: Callable,
) -> None:
    base_path = Path(__file__).parent
    filename = base_path / f"test_plot_actual_{lm_type}.jpg"
    plt.style.use("seaborn-v0_8")
    lmdiag.style.use("black_and_red")

    x = y = None
    if lm_type == "statsmodels":
        lm_fitted = statsmodels_factory(x_dims=5)
    elif lm_type == "linearmodels":
        lm_fitted = linearmodels_factory(x_dims=5)
    elif lm_type == "sklearn":
        lm_fitted, x, y = sklearn_factory(x_dims=5)
    else:
        raise ValueError(f"Unsupported lm_type: {lm_type}")

    # Set a specific font to make the test deterministic across different systems
    plt.rcParams["font.family"] = "sans-serif"
    plt.rcParams["font.sans-serif"] = ["Liberation Sans"]

    # Using lowess_delta leads to small differences between plot of
    # linearmodels and statsmodels, therefore we set it to 0 during testing
    fig = lmdiag.plot(lm_fitted, x=x, y=y, lowess_delta=0)
    fig.savefig(filename)

    # Log used font name, as different fonts lead to different jpg file sizes
    font_name = fig.get_axes()[0].title.get_fontproperties().get_name()

    # Quick and simple way to compare the plots is to compare jpg file sizes
    actual_bytes = Path(filename).stat().st_size
    expected_bytes = Path(base_path / "test_plot_expected.jpg").stat().st_size
    acceptable_byte_difference = 70
    assert abs(actual_bytes - expected_bytes) < acceptable_byte_difference, font_name



"""Module for diagnosis plots of linear regression models."""

from typing import Any, Optional

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import statsmodels.api as sm
from statsmodels.nonparametric.smoothers_lowess import lowess

from lmdiag import style
from lmdiag.statistics.base import StatsBase
from lmdiag.statistics.select import get_stats

LOWESS_DELTA = 0.005
LOWESS_IT = 2


def _get_figure(
    ax: Optional[mpl.axes.Axes] = None,
) -> tuple[mpl.figure.Figure, mpl.axes.Axes]:
    """Retrieve figure of axes or create a new one."""
    if ax is None:
        fig, ax = plt.subplots(**style.subplots)
    else:
        fig = ax.get_figure()
    return fig, ax


def resid_fit(
    lm: Any,
    x: Optional[np.ndarray] = None,
    y: Optional[np.ndarray] = None,
    ax: Optional[mpl.axes.Axes] = None,
    lowess_delta: float = LOWESS_DELTA,
    lowess_it: int = LOWESS_IT,
) -> mpl.figure.Figure:
    """Draw Residuals vs. Fitted Values Plot.

    For detailed explanation fo the lowess parameters, see:
    https://www.statsmodels.org/stable/generated/statsmodels.nonparametric.smoothers_lowess.lowess.html

    Args:
        lm: A fitted linear model of a supported type.
        x: X (predictor) of training data. Only for `sklearn` models!
        y: y (true response) of training data. Only for `sklearn` models!
        ax: Matplotlib axes for drawing. If `None` (default), a new Figure and Axes are
            created.
        lowess_delta: A value between 0 and 1. Higher values speed up plotting, but
            reduce the accuracy of the smoothing line. Defaults to 0.005. See:
        lowess_it: Lower values speed up plotting, but reduce the accuracy of the
            smoothing line. Defaults to 2.

    Returns:
        Figure of the plot.
    """
    lm_stats = lm if isinstance(lm, StatsBase) else get_stats(lm, x=x, y=y)

    fitted = lm_stats.fitted_values
    residuals = lm_stats.residuals
    smooth_x, smooth_y = lowess(residuals, fitted, it=lowess_it, delta=lowess_delta).T
    top_3_observations = np.abs(residuals).argsort()[-3:][::1]

    fig, ax = _get_figure(ax=ax)
    ax.plot([fitted.min(), fitted.max()], [0, 0], **style.plot_contour)
    ax.plot(smooth_x, smooth_y, **style.plot)
    ax.scatter(fitted, residuals, **style.scatter)
    for ob in top_3_observations:
        ax.annotate(ob, xy=(fitted[ob], residuals[ob]), **style.annotate)
    ax.set_title("Residual vs. Fitted", **style.title)
    ax.set_xlabel("Fitted values", **style.xy_label)
    ax.set_ylabel("Residuals", **style.xy_label)

    return fig


def q_q(
    lm: Any,
    x: Optional[np.ndarray] = None,
    y: Optional[np.ndarray] = None,
    ax: Optional[mpl.axes.Axes] = None,
) -> mpl.figure.Figure:
    """Draw Q-Q-Plot.

    Args:
        lm: A fitted linear model of a supported type.
        x: X (predictor) of training data. Only for `sklearn` models!
        y: y (true response) of training data. Only for `sklearn` models!
        ax: Matplotlib axes for drawing. If `None` (default), a new Figure and Axes are
            created.

    Returns:
        Figure of the plot.
    """
    lm_stats = lm if isinstance(lm, StatsBase) else get_stats(lm, x=x, y=y)

    std_resid = lm_stats.standard_residuals
    quantiles = lm_stats.normalized_quantiles
    std_resid_sorted = np.sort(std_resid)
    quantiles_sorted = np.sort(quantiles)
    fitted_values = np.polyfit(quantiles_sorted, std_resid_sorted, deg=1)
    fitted_line_y = fitted_values[0] * quantiles_sorted + fitted_values[1]

    # Index of sorted obs is used for annotation position, original index as label:
    top_3_sorted = np.abs(std_resid_sorted).argsort()[-3:][::1]
    top_3_orig = np.abs(std_resid).argsort()[-3:][::1]
    top_3_observations = zip(top_3_sorted, top_3_orig)

    fig, ax = _get_figure(ax=ax)
    ax.plot(quantiles_sorted, fitted_line_y, ":", **style.plot)
    ax.scatter(quantiles_sorted, std_resid_sorted, **style.scatter)
    for ob in top_3_observations:
        ax.annotate(
            ob[1],
            xy=(quantiles_sorted[ob[0]], std_resid_sorted[ob[0]]),
            **style.annotate,
        )
    ax.set_title("Normal Q-Q", **style.title)
    ax.set_xlabel("Theoretical Quantiles", **style.xy_label)
    ax.set_ylabel("Standardized residuals", **style.xy_label)

    return fig


def scale_loc(
    lm: Any,
    x: Optional[np.ndarray] = None,
    y: Optional[np.ndarray] = None,
    ax: Optional[mpl.axes.Axes] = None,
    lowess_delta: float = LOWESS_DELTA,
    lowess_it: int = LOWESS_IT,
) -> mpl.figure.Figure:
    """Draw Scale-Location Plot.

    For detailed explanation fo the lowess parameters, see:
    https://www.statsmodels.org/stable/generated/statsmodels.nonparametric.smoothers_lowess.lowess.html

    Args:
        lm: A fitted linear model of a supported type.
        x: X (predictor) of training data. Only for `sklearn` models!
        y: y (true response) of training data. Only for `sklearn` models!
        ax: Matplotlib axes for drawing. If `None` (default), a new Figure and Axes are
            created.
        lowess_delta: A value between 0 and 1. Higher values speed up plotting, but
            reduce the accuracy of the smoothing line. Defaults to 0.005. See:
        lowess_it: Lower values speed up plotting, but reduce the accuracy of the
            smoothing line. Defaults to 2.

    Returns:
        Figure of the plot.
    """
    lm_stats = lm if isinstance(lm, StatsBase) else get_stats(lm, x=x, y=y)

    fitted_vals = lm_stats.fitted_values
    sqrt_abs_res = lm_stats.sqrt_abs_residuals
    top_3_observations = sqrt_abs_res.argsort()[-3:][::1]
    smooth_x, smooth_y = lowess(
        sqrt_abs_res, fitted_vals, it=lowess_it, delta=lowess_delta
    ).T

    fig, ax = _get_figure(ax=ax)
    ax.plot(smooth_x, smooth_y, **style.plot)
    ax.scatter(fitted_vals, sqrt_abs_res, **style.scatter)
    for ob in top_3_observations:
        ax.annotate(ob, xy=(fitted_vals[ob], sqrt_abs_res[ob]), **style.annotate)
    ax.set_title("Scale-Location", **style.title)
    ax.set_xlabel("Fitted values", **style.xy_label)
    ax.set_ylabel(
        r"$\sqrt{\left|\mathregular{Standardized\ residuals}\right|}$", **style.xy_label
    )

    return fig


def resid_lev(
    lm: Any,
    x: Optional[np.ndarray] = None,
    y: Optional[np.ndarray] = None,
    ax: Optional[mpl.axes.Axes] = None,
    lowess_delta: float = LOWESS_DELTA,
    lowess_it: int = LOWESS_IT,
) -> mpl.figure.Figure:
    """Draw Standardized Residuals vs. Leverage Plot.

    For detailed explanation fo the lowess parameters, see:
    https://www.statsmodels.org/stable/generated/statsmodels.nonparametric.smoothers_lowess.lowess.html

    Args:
        lm: A fitted linear model of a supported type.
        x: X (predictor) of training data. Only for `sklearn` models!
        y: y (true response) of training data. Only for `sklearn` models!
        ax: Matplotlib axes for drawing. If `None` (default), a new Figure and Axes are
            created.
        lowess_delta: A value between 0 and 1. Higher values speed up plotting, but
            reduce the accuracy of the smoothing line. Defaults to 0.005. See:
        lowess_it: Lower values speed up plotting, but reduce the accuracy of the
            smoothing line. Defaults to 2.

    Returns:
        Figure of the plot.
    """
    lm_stats = lm if isinstance(lm, StatsBase) else get_stats(lm, x=x, y=y)

    std_resid = lm_stats.standard_residuals
    cooks_d = lm_stats.cooks_d
    leverage = lm_stats.leverage
    top_3_observations = cooks_d.argsort()[-3:][::1]
    x_ = np.linspace(leverage.min(), leverage.max(), 100)
    params_count = lm_stats.parameter_count
    smooth_x, smooth_y = lowess(std_resid, leverage, it=lowess_it, delta=lowess_delta).T

    fig, ax = _get_figure(ax=ax)

    # Draw cooks distance contours
    ax.plot(x_, np.sqrt((0.5 * params_count * (1 - x_)) / x_), **style.plot_contour)
    ax.plot(x_, np.sqrt((1.0 * params_count * (1 - x_)) / x_), **style.plot_contour)
    ax.plot(
        x_,
        np.negative(np.sqrt((0.5 * params_count * (1 - x_)) / x_)),
        **style.plot_contour,
    )
    ax.plot(
        x_,
        np.negative(np.sqrt((1.0 * params_count * (1 - x_)) / x_)),
        **style.plot_contour,
    )

    ax.plot(smooth_x, smooth_y, **style.plot)
    ax.scatter(leverage, std_resid, **style.scatter)

    # Limit y axis to actual values (otherwise contour lines disturb scale)
    ax.set_ylim(std_resid.min() * 1.1, std_resid.max() * 1.1)

    for ob in top_3_observations:
        ax.annotate(ob, xy=(leverage[ob], std_resid[ob]), **style.annotate)

    ax.set_title("Residuals vs. Leverage", **style.title)
    ax.set_xlabel("Leverage", **style.xy_label)
    ax.set_ylabel("Standardized residuals", **style.xy_label)

    return fig


def plot(
    lm: Any,
    x: Optional[np.ndarray] = None,
    y: Optional[np.ndarray] = None,
    lowess_delta: float = LOWESS_DELTA,
    lowess_it: int = LOWESS_IT,
) -> mpl.figure.Figure:
    """Plot all 4 charts as a Matrix.

    For detailed explanation fo the lowess parameters, see:
    https://www.statsmodels.org/stable/generated/statsmodels.nonparametric.smoothers_lowess.lowess.html

    Args:
        lm: A fitted linear model of a supported type.
        x: X (predictor) of training data. Only for `sklearn` models!
        y: y (true response) of training data. Only for `sklearn` models!
        lowess_delta: A value between 0 and 1. Higher values speed up plotting, but
            reduce the accuracy of the smoothing line. Defaults to 0.005. See:
        lowess_it: Lower values speed up plotting, but reduce the accuracy of the
            smoothing line. Defaults to 2.

    Returns:
        Figure of the plot.
    """
    lm_stats = get_stats(lm=lm, x=x, y=y)

    fig, axs = plt.subplots(2, 2, **style.subplots)

    resid_fit(lm_stats, ax=axs[0][0], lowess_delta=lowess_delta, lowess_it=lowess_it)
    q_q(lm_stats, ax=axs[0][1])
    scale_loc(lm_stats, ax=axs[1][0], lowess_delta=lowess_delta, lowess_it=lowess_it)
    resid_lev(lm_stats, ax=axs[1][1], lowess_delta=lowess_delta, lowess_it=lowess_it)

    fig.tight_layout(**style.tight_layout)

    return fig


if __name__ == "__main__":
    # Example used for debugging
    df = sm.datasets.get_rdataset("ames", "openintro").data
    y = np.log10(df["price"])
    X = df["Overall.Qual"] + np.log(df["area"])
    X = sm.add_constant(X)

    lm = sm.OLS(y, X).fit()
    fig = plot(lm)
    fig.savefig("test.png")



from typing import Union

import numpy as np
import statsmodels.api as sm
from statsmodels.stats.outliers_influence import OLSInfluence, summary_table

from lmdiag.statistics.base import StatsBase, optionally_cached_property


class StatsmodelsStats(StatsBase):
    def __init__(
        self,
        lm: sm.regression.linear_model.RegressionResultsWrapper,
        cache: bool = True,
    ) -> None:
        super().__init__()
        self._lm = lm
        self._cache_properties = cache
        self.__ols_influence: Union[OLSInfluence, None] = None

    @property
    def _ols_influence(self) -> OLSInfluence:
        if self.__ols_influence is None:
            self.__ols_influence = OLSInfluence(self._lm)
        return self.__ols_influence

    @optionally_cached_property
    def residuals(self) -> np.ndarray:
        _, data, _ = summary_table(self._lm, alpha=0.05)
        return data[:, 8]

    @optionally_cached_property
    def fitted_values(self) -> np.ndarray:
        return self._lm.fittedvalues

    @optionally_cached_property
    def standard_residuals(self) -> np.ndarray:
        return self._ols_influence.resid_studentized_internal

    @optionally_cached_property
    def cooks_d(self) -> np.ndarray:
        return self._ols_influence.cooks_distance[0]

    @optionally_cached_property
    def leverage(self) -> np.ndarray:
        return self._ols_influence.hat_matrix_diag

    @optionally_cached_property
    def parameter_count(self) -> int:
        return len(self._lm.params)
